"""
Module for creating a simple retrieval-augmented generation (RAG) graph using LangChain.

© 2024 Blaine Freestone, Carson Bush

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from langchain_core.documents.base import Document
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langgraph.graph import StateGraph
from langchain_core.output_parsers import StrOutputParser
from langchain_openai import ChatOpenAI
from langgraph.graph.graph import CompiledGraph
from typing_extensions import TypedDict
from typing import List, Annotated
from langchain_core.vectorstores import VectorStoreRetriever
from langchain_community.vectorstores import FAISS
from langchain_openai import OpenAIEmbeddings
from langgraph.checkpoint.sqlite import SqliteSaver

def get_simple_rag(
    vectorstore_path: str,
    vectorstore_index: str,
    memory_filepath: str,
    api_key: str | None = None,
    system_prompt_text: str = (
        'You are a helpful teacher helping a student with course material.\n'
        'You will answer a question based on the context provided:\n'
        'Don\'t answer questions about other things.\n\n'
        '{context}\n'
    ),
    model: str = 'gpt-4o-mini'
) -> CompiledGraph:
    """Create a simple retrieval-augmented generation (RAG) graph.
    
    Args:
        vectorstore_path (str): Path to the vector store.
        vectorstore_index (str): Index name for the vector store.
        memory_filepath (str): Filepath for the memory checkpoint.
        api_key (str | None): API key for the language model. Defaults to None.
        system_prompt_text (str): Prompt text for the system message. Defaults to a helpful teacher prompt.
        model (str): Model name for the language model. Defaults to 'gpt-4o-mini'.
    
    Returns:
        CompiledGraph: The compiled state graph.
    """

    def add_messages(left: list, right: list):
        """Add-don't-overwrite."""
        return left + right

    class GraphState(TypedDict):
        """Represents the state of the graph."""
        retrieved_context: List[Document]
        messages: Annotated[list, add_messages]

    llm: ChatOpenAI = ChatOpenAI(model=model) if api_key is None else ChatOpenAI(api_key=api_key, model=model)  # type: ignore

    retriever: VectorStoreRetriever = FAISS.load_local(
        vectorstore_path,
        OpenAIEmbeddings() if api_key is None else OpenAIEmbeddings(api_key=api_key),  # type: ignore
        allow_dangerous_deserialization=True,
        index_name=vectorstore_index,
    ).as_retriever()

    system_prompt: ChatPromptTemplate = ChatPromptTemplate.from_messages([
        ('system', system_prompt_text),
        MessagesPlaceholder('messages'),
        ('human', '{input}'),
    ])

    chain = system_prompt | llm | StrOutputParser()

    def retrieve_node(state: GraphState) -> dict:
        """Retrieve context documents based on the latest question."""
        question = state['messages'][-1]
        documents: List[Document] = retriever.invoke(question)
        return {'retrieved_context': documents}

    def generate_node(state: GraphState) -> dict:
        """Generate a response based on the context and messages."""
        messages = state['messages']
        documents: List[Document] = state['retrieved_context']
        generation: str = chain.invoke({
            'context': documents,
            'messages': messages[:-1],
            'input': messages[-1],
        })
        return {'messages': [generation]}

    graph = StateGraph(GraphState)

    graph.add_node('retrieve', retrieve_node)
    graph.add_node('generate', generate_node)
    graph.add_edge('retrieve', 'generate')
    graph.set_entry_point('retrieve')
    graph.set_finish_point('generate')

    memory = SqliteSaver.from_conn_string(f'{memory_filepath}')
    compiled_graph: CompiledGraph = graph.compile(checkpointer=memory)
    return compiled_graph



"""
This module contains functions to save training data and handle controller
requests in a Flask application.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager
from flask import request, redirect

def controller(chat_session_manager: ChatSessionManager):
    """
    Handle the POST request to save training data.

    Args:
        chat_session_manager (ChatSessionManager): The chat session manager instance.
    """
    chat_logs_manager = chat_session_manager.chat_logs_manager
    
    name = request.form.get('name')
    role = request.form.get('role')
    type = request.form.get('type')
    question = request.form.get('question')
    answer: str | None = request.form.get('answer')

    if chat_logs_manager is not None:
        chat_logs_manager.save_training_data({
            'name': name,
            'role': role,
            'type': type,
            'question': question,
            'answer': answer
        })

    return redirect('/')



"""
This module contains functions for processing and rendering chat logs.

It includes functions to process messages, get the log file template, and display
the content of a specified log file.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager
from flask import abort

def controller(chat_sessions_manager: ChatSessionManager, branch, filename, app_name: str | None = None):
    """
    Display the content of a specified log file.

    Args:
        chat_sessions_manager (ChatSessionManager): The chat sessions manager instance.
        branch (str): The branch where the log file is located.
        filename (str): The name of the log file.
    
    Returns:
        str: Rendered template with log file content.
    
    Raises:
        FileNotFoundError: If the log file is not found.
        yaml.YAMLError: If there is an error parsing the log file.
    """
    chat_logs_manager = chat_sessions_manager.chat_logs_manager
    
    return chat_logs_manager.get_log_file_template(filename, branch) if chat_logs_manager else abort(404)



"""Module for handling feedback form display.

This module contains the controller function to render the feedback form template.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import render_template

def controller(app_name: str | None = None, main_logo_login: str | None = None, main_logo_chat: str | None = None, favicon: str | None = None) -> str:
    """
    Display the feedback form.

    Returns:
        str: Rendered feedback form template.
    """
    role_options = ['Undergraduate Student', 'Graduate Student', 'Faculty', 'Other']
    category_options = ['Other', 'General Feedback', 'Bug Report', 'Feature Request', 'Content Issue']
    return render_template(
        'feedback_form.html', 
        role_options=role_options,
        category_options=category_options,
        main_logo_login=main_logo_login,
        main_logo_chat=main_logo_chat,
        favicon=favicon,
        app_name=app_name if app_name else "Maeser"
    )



"""
This is the chat subpackage for the Maeser package.

This package contains the following subpackages and modules:

- `chat_logs`: This module provides functionality for managing chat logs.
- `chat_session_manager`: This module provides functionality for managing chat sessions.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from . import chat_logs
from . import chat_session_manager

__all__ = ["chat_logs", "chat_session_manager"]


"""
Module for creating a simple retrieval-augmented generation (RAG) graph using LangChain.

© 2024 Gohaun Manley

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""
from langchain_core.documents.base import Document
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langgraph.graph import StateGraph, START, END
from langchain_core.output_parsers import StrOutputParser
from langchain_openai import ChatOpenAI
from langchain_core.messages import SystemMessage
from langgraph.graph.graph import CompiledGraph
from typing_extensions import TypedDict
from typing import List, Dict, Annotated
from langchain_community.vectorstores import FAISS
from langchain_openai import OpenAIEmbeddings
from langgraph.checkpoint.sqlite import SqliteSaver

def add_messages(left: List[str], right: List[str]) -> List[str]:
    return left + right

class GraphState (TypedDict):
    messages: Annotated[list, add_messages]
    current_topic: str | None = None
    retrieved_context: List[Document] | None = None
    first_messsage: bool = True

def normalize_topic(topic: str) -> str:
    """
    Converts the input topic string to lowercase.
    """
    return topic.lower()

def remove_context_placeholder(prompt: str) -> str:
    """
    Remove the '{context}' placeholder from a prompt string.
    """
    return prompt.replace("{context}", "").strip()

def get_pipeline_rag (
    vectorstore_config: Dict[str, str],
    memory_filepath: str,
    api_key: str | None = None,
    system_prompt_text: str = (
        'You are a helpful teacher helping a student with course material.\n'
        'You will answer a question based on the context provided:\n'
        'Don\'t answer questions about other things.\n\n'
        '{context}\n'
    ),
    model: str = 'gpt-4o-mini'
) -> CompiledGraph:
    """
    Create a dynamic retrieval-augmented generation (RAG) graph that includes topic extraction,
    conditional routing to retrieval nodes, and answer generation. The returned object is a
    compiled graph (with memory checkpoint) that you can run by providing an initial state.
    
    Args:
        vectorstore_config (Dict[str, Tuple[str, str]]): Mapping of topic to (vectorstore_path, index name) *WARNING* TOPIC MUST BE ALL LOWER CASE
        memory_filepath (str): Path for the memory checkpoint (SQLite database).
        api_key (Optional[str]): API key for language models and embeddings.
        system_prompt_text (str): System prompt template for answer generation.
        model (str): Model name to use.
    
    Returns:
        CompiledGraph: A compiled state graph ready for execution.
    """

    # initalize FAISS retreivers for each topic 
    # (i.e load each vectorstore to be used when it is needed)
    retrievers = {}
    for topic, vstore_path in vectorstore_config.items():
        retrievers[topic] = FAISS.load_local(
            vstore_path,
            OpenAIEmbeddings() if api_key is None else OpenAIEmbeddings(api_key=api_key),
            allow_dangerous_deserialization=True
        ).as_retriever()

    # Build the Chain for the generate node
    system_prompt = ChatPromptTemplate.from_messages([
        ('system', system_prompt_text),
        MessagesPlaceholder('messages'),
        ('human', "{input}")
    ])
    llm = ChatOpenAI(model=model, temperature=0) if api_key is None else ChatOpenAI(api_key=api_key, model=model, temperature=0)
    chain = system_prompt | llm | StrOutputParser()

    #format topics for later topic extraction
    def format_topic_keys(topics):
        keys = list(topics.keys())  # Get dictionary keys as a list
        if not keys:
            return ""  # Return empty string if dictionary is empty
        elif len(keys) == 1:
            return f"'{keys[0]}'"  # If only one key, return it without "or"
        else:
            return ", ".join(f"'{key}'" for key in keys[:-1]) + f", or '{keys[-1]}'"

    def determine_topic_node (state: GraphState, vectorstore_config: Dict) -> dict:

        # Prepare the list of valid topics plus "off topic"
        formatted_topics = format_topic_keys(vectorstore_config)
        current_topic = state.get("current_topic")

        # Build a prompt that includes the current topic (if any) and the user message.
        clean_system_prompt = remove_context_placeholder(system_prompt_text)
        prompt_template = ChatPromptTemplate.from_messages([
            ("system", f"You are an assistant who extracts a concise topic label from a user's explanation. Here is the Current Topic: {current_topic}. If the Current Topic is None, please choose a valid topic."
                       f"The AI who will be answering the users questions and using your topics has been given this prompt {clean_system_prompt}. Please use this as part of your consideration for the topic."
                       f"Using these topics exactly ({formatted_topics}), if the user's latest message indicates that the topic should change, "
                       f"output the new topic; otherwise, repeat the current topic."),
            ("human", "User message: {question}\nExtract the topic:")
        ])

        question = state["messages"][-1]
        formatted_prompt = prompt_template.format(question=question, current_topic=current_topic if current_topic else "None")
        llm_topic = ChatOpenAI(model=model, temperature=0) if api_key is None else ChatOpenAI(api_key=api_key, model=model, temperature=0)
        result = llm_topic.invoke([SystemMessage(content=formatted_prompt)])
        topic = normalize_topic(result.content)
        return {"current_topic": topic}
    
    # Create a factory for retrieval nodes to return relevant information
    def make_retrieval_node(topic: str):
        def retrieval_node(state: GraphState) -> dict:
            question = state["messages"][-1]
            documents: List[Document] = retrievers[topic].invoke(question)
            return {"retrieved_context": documents}
        return retrieval_node
    
    # Map topics with retrieval node names
    vectorstore_nodes = {}
    for topic in vectorstore_config.keys():
        node_name = f"retrieve_{topic}"
        vectorstore_nodes[topic] = node_name

    # Node: answer generation.
    def generate_node(state: GraphState) -> dict:
        messages = state["messages"]
        documents = state.get("retrieved_context", [])
        generation = chain.invoke({
            "context": documents,
            "input": messages[-1],
            "messages": messages[:-1],
        })
        # Update conversation history with the generated answer.
        return {"messages": messages + [generation]}
    
    # Build the state graph.
    graph = StateGraph(GraphState)
    graph.add_node("determine_topic", lambda state: determine_topic_node(state, vectorstore_config))    

    # Set up conditional branching based on the determined topic.
    # Mapping: if "off topic", go to off_topic_response; if valid topic, go to its retrieval node.
    mapping = {}
    for topic in vectorstore_config.keys():
        mapping[topic] = f"retrieve_{topic}"
    graph.add_conditional_edges("determine_topic", lambda state: state["current_topic"], mapping)
    
    # Add retrieval nodes for each valid topic.
    for topic in vectorstore_config.keys():
        node_name = f"retrieve_{topic}"
        graph.add_node(node_name, make_retrieval_node(topic))
        graph.add_edge(node_name, "generate")
    
    # Add answer generation node.
    graph.add_node("generate", generate_node)
    
    # Define the overall flow.
    graph.add_edge(START, "determine_topic")
    graph.add_edge("generate", END)
    
    # Set up memory checkpoint using SQLite.
    memory = SqliteSaver.from_conn_string(f'{memory_filepath}')
    compiled_graph: CompiledGraph = graph.compile(checkpointer=memory)
    return compiled_graph


"""
Blueprint definitions for the Maeser application.

This module sets up the Flask blueprint and associated routes for the Maeser
application. It includes route handlers for chat interfaces, user management,
feedback, and training functionalities.

© 2024 Blaine Freestone, Carson Bush, Gohaun Manley

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import Blueprint, Flask
from flask_login import login_required, current_user, LoginManager
import os
from datetime import datetime
import threading
import time

from maeser.chat.chat_session_manager import ChatSessionManager
from maeser.controllers.common.decorators import admin_required, rate_limited
from maeser.user_manager import UserManager
from .controllers import (
    chat_api,
    chat_interface,
    chat_logs_overview,
    display_chat_log,
    feedback_api,
    feedback_form_get,
    feedback_form_post,
    login_api,
    logout,
    new_session_api,
    training,
    training_post,
    conversation_history_api,
    remaining_requests_api,
    manage_users_view,
    user_management_api,
)


class App_Manager:
    """
    Manages the Maeser App and its configurations.

    Args:
        app (Flask): The Flask application instance.
        app_name (str | None, optional): The name of the application.
        flask_secret_key (str): The secret key for the Flask application.
        chat_session_manager (ChatSessionManager): The chat session manager instance.
        user_manager (UserManager | None, optional): The user manager instance. Defaults to None.
        main_logo_login (str | None, optional): URL or path to the login logo. Defaults to None.
        main_logo_chat (str | None, optional): URL or path to the chat logo. Defaults to None.
        chat_head (str | None, optional): URL or path to the chat header image. Defaults to None.
        favicon (str | None, optional): URL or path to the favicon. Defaults to None.
        login_text (str | None, optional): Text to display on the login page. Defaults to None.
        changelog (str | None, optional): Text to display as the changelog. Defaults to None.
        chat_greeting (str, optional): Greeting message to display in the chat. Defaults to "Hello, how can I help you today?".
        branch_response (str, optional): Response message to display when a branch is selected. Defaults to "Okay, I'll help you with ${action}!".
        animation (bool, optional): Whether to enable toggle animations. Defaults to False.
        primary_color (str, optional): Primary color of the application. Defaults to "#333".
        secondary_color (str, optional): Secondary color of the application. Defaults to "#ccc".
        button_color (str, optional): Color of the buttons. Defaults to "#0084ff".
        button_color_active (str, optional): Color of the active buttons. Defaults to "#009e15".
        button_color_inactive (str, optional): Color of the inactive buttons. Defaults to "#ddd".
        fafa_button (str, optional): Color of the fafa buttons. Defaults to "#eee".
        logout_button (str, optional): Color of the logout button. Defaults to "#333".
        new_chat_button (str, optional): Color of the new chat button. Defaults to "#333".
        help_train_button (str, optional): Color of the help train button. Defaults to "#eee".
    """

    def __init__(
        self,
        # app and app name
        app: Flask,
        app_name: str,
        # flask secret
        flask_secret_key: str,
        # managers
        chat_session_manager: ChatSessionManager,
        user_manager: UserManager | None = None,
        # images
        main_logo_login: str | None = None,
        main_logo_chat: str | None = None,
        chat_head: str | None = None,
        favicon: str | None = None,
        # text
        login_text: str | None = None,
        changelog: str | None = None,
        chat_greeting: str = "Hello, how can I help you today?",
        branch_response: str = "Okay, I'll help you with ${action}!",
        # toggle animations
        animation: bool = False,
        # colors
        primary_color: str = "#f5f5f5",
        secondary_color: str = "#ccc",
        button_color: str = "#0084ff",
        button_color_active: str = "#009e15",
        button_color_inactive: str = "#e9e9e9",
        fafa_button: str = "#b8b8b8",
        logout_button: str = "#000000",
        new_chat_button: str = "#000000",
        help_train_button: str = "#a8a8a8",
    ):
        self.app = app
        self.app_name = app_name

        self.flask_secret_key = flask_secret_key

        self.chat_session_manager = chat_session_manager
        self.user_manager = user_manager

        self.main_logo_login = main_logo_login
        self.main_logo_chat = main_logo_chat
        self.chat_head = chat_head
        self.favicon = favicon

        self.login_text = login_text
        self.changelog = changelog
        self.chat_greeting = chat_greeting
        self.branch_response = branch_response

        self.animation = animation

        self.primary_color = primary_color
        self.secondary_color = secondary_color
        self.button_color = button_color
        self.button_color_active = button_color_active
        self.button_color_inactive = button_color_inactive
        self.fafa_button = fafa_button
        self.logout_button = logout_button
        self.new_chat_button = new_chat_button
        self.help_train_button = help_train_button

        self.current_dir = os.path.dirname(os.path.abspath(__file__ + "/."))

        # The following functions with no code are work in progress and will be added soon

    def template_styles_css(self):
        """
        Template the styles.css file using Jinja2 templating
        """

    def template_chat_interface(self):
        """
        Template the chat_interface.html file using Jinja2 templating
        """

    def template_login(self):
        """
        Template the login.html file using Jinja2 templating
        """

    def template_training(self):
        """
        Template the training.html file using Jinja2 templating
        """

    def template_feedback(self):
        """
        Template the feedback_form.html file using Jinja2 templating
        """

    def add_flask_blueprint(self) -> Flask:
        """
        Add the Maeser blueprint to the Flask application.

        Returns:
            Flask: The Flask application instance with the blueprint registered.
        """
        maeser_blueprint = Blueprint(
            "maeser",
            __name__,
            template_folder=os.path.join(self.current_dir, "data/templates"),
            static_folder=os.path.join(self.current_dir, "data/static"),
            static_url_path="/maeser/static",
        )

        @maeser_blueprint.app_template_filter("datetimeformat")
        def datetimeformat(value, format="%Y-%m-%d %H:%M:%S"):
            """Format a timestamp as a datetime string."""
            return datetime.fromtimestamp(value).strftime(format)

        if self.user_manager:

            def refresh_requests():
                """Refresh user requests at regular intervals."""
                while True:
                    time.sleep(self.user_manager.rate_limit_interval)
                    self.user_manager.refresh_requests()

            threading.Thread(target=refresh_requests, daemon=True).start()

            self.app.secret_key = self.flask_secret_key
            login_manager = LoginManager(self.app)
            login_manager.init_app(self.app)
            login_manager.login_view = "maeser.login"  # type: ignore
            login_manager.session_protection = "strong"

            @login_manager.user_loader
            def load_user(user_full_id: str):
                """Load a user by their ID."""
                auth_method, user_id = user_full_id.split(".", 1)
                return self.user_manager.get_user(auth_method, user_id)

            @maeser_blueprint.route("/")
            @login_required
            def chat_interface_route():
                """Route for the chat interface."""
                return chat_interface.controller(
                    self.chat_session_manager,
                    self.user_manager.max_requests,
                    self.user_manager.rate_limit_interval,
                    current_user,
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    chat_head=self.chat_head,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/login", methods=["GET", "POST"])
            def login():
                """Route for the login interface."""
                return login_api.login_controller(
                    self.user_manager,
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/login/github", methods=["GET"])
            def github_authorize():
                """Route for GitHub authorization."""
                return login_api.github_authorize_controller(
                    current_user, self.user_manager.authenticators["github"]
                )

            @maeser_blueprint.route("/login/github_callback")
            def github_auth_callback():
                """Route for GitHub authorization callback."""
                return login_api.github_auth_callback_controller(
                    current_user,
                    self.user_manager,
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/logout")
            @login_required
            def logout_route():
                """Route for logging out."""
                return logout.controller()

            @maeser_blueprint.route("/users")
            @login_required
            @admin_required(current_user)
            def manage_users():
                """Route for managing users."""
                return manage_users_view.controller(
                    self.user_manager,
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/users/api", methods=["POST"])
            @login_required
            @admin_required(current_user)
            def manage_users_api():
                """API route for managing users."""
                return user_management_api.controller(self.user_manager)

            @maeser_blueprint.route("/req_session", methods=["POST"])
            @login_required
            def sess_handler():
                """Route for handling session requests."""
                return new_session_api.controller(self.chat_session_manager, True)

            @maeser_blueprint.route("/msg/<chat_session>", methods=["POST"])
            @login_required
            @rate_limited(self.user_manager, current_user)
            def msg_api(chat_session):
                """API route for handling chat messages."""
                return chat_api.controller(self.chat_session_manager, chat_session)

            @maeser_blueprint.route("/feedback", methods=["POST"])
            @login_required
            def feedback():
                """Route for submitting feedback."""
                return feedback_api.controller(self.chat_session_manager)

            @self.app.route("/get_requests_remaining", methods=["GET"])
            @login_required
            def get_requests_remaining():
                """Route for getting the remaining requests."""
                return remaining_requests_api.controller(
                    self.user_manager, current_user
                )

        else:

            @maeser_blueprint.route("/")
            def chat_interface_route():
                """Route for the chat interface."""
                return chat_interface.controller(
                    self.chat_session_manager,
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    chat_head=self.chat_head,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/req_session", methods=["POST"])
            def sess_handler():
                """Route for handling session requests."""
                return new_session_api.controller(self.chat_session_manager)

            @maeser_blueprint.route("/msg/<chat_session>", methods=["POST"])
            def msg_api(chat_session):
                """API route for handling chat messages."""
                return chat_api.controller(self.chat_session_manager, chat_session)

            @maeser_blueprint.route("/feedback", methods=["POST"])
            def feedback():
                """Route for submitting feedback."""
                return feedback_api.controller(self.chat_session_manager)

        if self.chat_session_manager.chat_logs_manager:

            @maeser_blueprint.route("/train")
            @login_required if self.user_manager else lambda x: x
            def train():
                """Route for training."""
                return training.controller(
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/submit_train", methods=["POST"])
            @login_required if self.user_manager else lambda x: x
            def submit_train():
                """Route for submitting training data."""
                return training_post.controller(self.chat_session_manager)

            @maeser_blueprint.route("/feedback_form")
            def feedback_form():
                """Route for getting the feedback form."""
                return feedback_form_get.controller(
                    main_logo_login=self.main_logo_login,
                    main_logo_chat=self.main_logo_chat,
                    favicon=self.favicon,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/submit_feedback", methods=["POST"])
            def submit_feedback():
                """Route for submitting feedback."""
                return feedback_form_post.controller(self.chat_session_manager)

            @maeser_blueprint.route("/logs", methods=["GET"])
            @login_required if self.user_manager else lambda x: x
            @admin_required(current_user) if self.user_manager else lambda x: x
            def logs():
                """Route for viewing chat logs."""
                return chat_logs_overview.controller(
                    self.chat_session_manager,
                    favicon=self.favicon,
                    app_name=self.app_name,
                )

            @maeser_blueprint.route("/logs/<branch>/<filename>")
            @login_required if self.user_manager else lambda x: x
            @admin_required(current_user) if self.user_manager else lambda x: x
            def display_log(branch, filename):
                """Route for displaying a specific chat log."""
                return display_chat_log.controller(
                    self.chat_session_manager, branch, filename, app_name=self.app_name
                )

            @maeser_blueprint.route("/conversation_history", methods=["POST"])
            @login_required if self.user_manager else lambda x: x
            def conversation_history():
                """Route for getting conversation history."""
                return conversation_history_api.controller(self.chat_session_manager)

        self.app.register_blueprint(maeser_blueprint)
        return self.app



"""
This is the Maeser package, which provides a set of classes and
functions for managing a chat application.

The package is organized as follows:

- `chat`: This module contains classes and functions related to chat functionality,
          such as sending and receiving messages.
- `controllers`: This module contains classes and functions for managing Flask
          controllers, which handle the logic for handling incoming requests.
- `user_manager`: This module contains classes and functions for managing users
          in the chat application.
- `render`: This module contains classes and functions for rendering the user
          interface of the chat application.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from . import chat
from . import controllers
from . import user_manager
from . import render

__all__ = ['chat', 'controllers', 'user_manager', 'render']



"""
This module contains decorators for rate limiting and admin access control
in a Flask application.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from functools import wraps
from flask import abort

def rate_limited(auth_manager, current_user):
    """
    Decorator to rate limit an endpoint based on user's remaining requests.

    Args:
        auth_manager: The authentication manager to handle request limits.
        current_user: The user object containing request information.

    Returns:
        A wrapped endpoint function that checks for rate limits.
    """
    def decorator(endpoint):
        @wraps(endpoint)
        def rate_limited_wrapper(*args, **kwargs):
            # Check if user has any requests remaining before proceeding
            if current_user.requests_remaining <= 0:
                print(f'User ({current_user.full_id_name}) has no requests remaining')
                abort(429, 'Rate limit reached, please try again later')

            result = endpoint(*args, **kwargs)

            # Decrease the number of requests remaining for the user once response is sent and update the result
            auth_manager.decrease_requests(current_user.auth_method, current_user.ident)
            result['requests_remaining'] = auth_manager.get_requests_remaining(current_user.auth_method, current_user.ident)
            return result
        
        rate_limited_wrapper.__name__ = f'{endpoint.__name__}'

        return rate_limited_wrapper
    return decorator

def admin_required(current_user):
    """
    Decorator to ensure that an endpoint can only be accessed by an admin.

    Args:
        current_user: The user object to check for admin privileges.

    Returns:
        A wrapped endpoint function that checks for admin access.
    """
    def decorator(endpoint):
        @wraps(endpoint)
        def admin_wrapper(*args, **kwargs):
            if current_user.admin:
                print(f'User ({current_user.full_id_name}) is admin')
                return endpoint(*args, **kwargs)
            print(f'User ({current_user.full_id_name}) is not authorized')
            abort(403, 'Admin access is required to access this page.')
        
        admin_wrapper.__name__ = f'{endpoint.__name__}'
        
        return admin_wrapper
    return decorator



"""
Module for managing chat logs, including logging and retrieving chat history,
feedback, and training data.

© 2024 Blaine Freestone, Carson Bush

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.user_manager import UserManager, User
from maeser.render import get_response_html
from abc import ABC, abstractmethod
from datetime import datetime
import time
import yaml
from os import path, stat, walk, mkdir, makedirs
import subprocess
from flask import abort, render_template
import platform


class BaseChatLogsManager(ABC):
    def __init__(
        self, chat_log_path: str, user_manager: UserManager | None = None
    ) -> None:
        """
        Initializes the BaseChatLogsManager.

        Args:
            chat_log_path (str): Path to the chat log directory.
            user_manager (UserManager | None): Optional user manager instance.
        """
        self.chat_log_path: str = chat_log_path
        self.user_manager: UserManager | None = user_manager

        # create log directory if it does not exist
        if not path.exists(self.chat_log_path):
            makedirs(self.chat_log_path)

    @abstractmethod
    def log(self, branch_name: str, session_id: str, log_data: dict) -> None:
        """
        Abstract method to log chat data.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            log_data (dict): The data to be logged.

        Returns:
            None
        """
        pass

    @abstractmethod
    def log_feedback(
        self, branch_name: str, session_id: str, message_index: int, feedback: str
    ) -> None:
        """
        Abstract method to log feedback for a message.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            message_index (int): The index of the message to add feedback to.
            feedback (str): The feedback to add to the message.

        Returns:
            None
        """
        pass

    @abstractmethod
    def get_chat_history_overview(self, user: User | None) -> list[dict]:
        """
        Abstract method to get an overview of chat history.
        This is used to display a list of overviews of previous chat conversations.

        Args:
            user (User | None): The user to get chat history for.

        Returns:
            list[dict]: A list of dictionaries containing information about previous chat conversations. Each should have the following keys:
                - 'branch': The name of the branch.
                - 'session': The session ID for the conversation.
                - 'modified': The timestamp of when the chat conversation was last modified.
                - 'header': The text that will be used as the link text. Usually the first message in the conversation. Could also be a conversation title.
        """
        pass

    @abstractmethod
    def get_chat_logs_overview(
        self, sort_by: str, order: str, branch_filter: str, user_filter: str, feedback_filter: str
    ) -> tuple[list[dict], int, float, set[str]]:
        """
        Abstract method to get an overview of chat logs.

        Args:
            sort_by (str): The field to sort by.
            order (str): The order to sort by. Either 'asc' or 'desc'.
            branch_filter (str): The branch to filter by.
            feedback_filter (str): The feedback to filter by.

        Returns:
            tuple: A tuple containing:
                - list[dict]: A list of dictionaries containing information about chat logs.
                - int: The total number of tokens used.
                - float: The total cost of the chat logs.
        """
        pass

    @abstractmethod
    def get_chat_history(self, branch_name: str, session_id: str) -> dict:
        """
        Abstract method to get chat history for a session.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.

        Returns:
            dict: The chat history for the session.
        """
        pass

    @abstractmethod
    def get_log_file_template(self, filename: str, branch: str) -> str:
        """
        Abstract method to get the jinja template for a log file.

        Args:
            filename (str): The name of the log file.
            branch (str): The branch the log file is in.

        Returns:
            str: The rendered template for the log file.
        """
        pass

    @abstractmethod
    def save_feedback(self, feedback: dict) -> None:
        """
        Abstract method to save feedback input to a file.

        Args:
            feedback (dict): The feedback to save.

        Returns:
            None
        """
        pass

    @abstractmethod
    def save_training_data(self, training_data: dict) -> None:
        """
        Abstract method to save training data to a file.

        Args:
            training_data (dict): The training data to save.

        Returns:
            None
        """
        pass


class ChatLogsManager(BaseChatLogsManager):
    def __init__(self, chat_log_path: str) -> None:
        """
        Initializes the ChatLogsManager.

        Args:
            chat_log_path (str): Path to the chat log directory.
        """
        super().__init__(chat_log_path)

    def log(self, branch_name: str, session_id: str, log_data: dict) -> None:
        """
        Logs chat data to a YAML file.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            log_data (dict): The data to be logged. Should contain the following keys: 'user', 'cost', 'tokens', and 'message'.

        Returns:
            None
        """
        if not self._does_log_exist(branch_name, session_id):
            self._create_log_file(branch_name, session_id, log_data.get("user", None))
        else:
            self._update_log_file(branch_name, session_id, log_data)

    def log_feedback(
        self, branch_name: str, session_id: str, message_index: int, feedback: str
    ) -> None:
        """
        Adds feedback to the log for a specific response in a specific session.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            message_index (int): The index of the message to add feedback to.
            feedback (str): The feedback to add to the message.

        Returns:
            None
        """
        with open(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log", "r"
        ) as file:
            log: dict = yaml.safe_load(file)
            log["messages"][message_index]["liked"] = feedback

        with open(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log", "w"
        ) as file:
            yaml.dump(log, file)

    def get_chat_history_overview(self, user: User | None) -> list[dict]:
        """
        Gets an overview of chat history.

        Args:
            user (User | None): The user to get chat history for.

        Returns:
            list[dict]: A list of dictionaries containing information about previous chat conversations.
        """
        overview = []
        conversations = self._get_file_list()
        for conversation in conversations:
            current_user_name: str = "anon" if user is None else user.full_id_name
            if current_user_name == conversation["user"]:
                overview.append(
                    {
                        "branch": conversation["branch"],
                        "session": conversation["name"].removesuffix(".log"),
                        "modified": conversation["modified"],
                        "header": conversation["first_message"],
                    }
                )
        # Sort conversations by date modified
        overview.sort(key=lambda x: x["modified"], reverse=True)

        # Remove conversations with no first message
        overview = [link for link in overview if link["header"] is not None]

        return overview

    def get_chat_logs_overview(
        self, sort_by: str, order: str, branch_filter: str, user_filter: str, feedback_filter: str
    ) -> tuple[list[dict], int, float, set[str]]:
        """
        Gets an overview of chat logs.

        Args:
            sort_by (str): The field to sort by.
            order (str): The order to sort by. Either 'asc' or 'desc'.
            branch_filter (str): The branch to filter by.
            feedback_filter (str): The feedback to filter by.

        Returns:
            tuple: A tuple containing:
                - list[dict]: A list of dictionaries containing information about chat logs.
                - int: The total number of tokens used.
                - float: The total cost of the chat logs.
        """
        log_files = self._get_file_list()

        # Get set of all users
        user_set = {f["user"] for f in log_files}

        if branch_filter:
            log_files = [
                f for f in log_files if branch_filter.lower() in f["branch"].lower()
            ]

        if user_filter:
            log_files = [
                f for f in log_files if user_filter == f["user"]
            ]

        if feedback_filter:
            feedback_filter_bool = feedback_filter.lower() == "true"
            log_files = [
                f for f in log_files if f["has_feedback"] == feedback_filter_bool
            ]

        reverse = order == "desc"
        log_files.sort(key=lambda x: x[sort_by], reverse=reverse)

        # Calculate aggregate number of tokens and cost
        total_tokens = 0
        total_cost = 0.0
        for file in log_files:
            with open(
                f'{self.chat_log_path}/chat_history/{file["branch"]}/{file["name"]}',
                "r",
            ) as f:
                log = yaml.safe_load(f)
                if log.get("total_tokens") is None:
                    print(
                        f"\x1b[33mWarning: \"total_tokens\" key is missing from log for file {file['name']}, defaulting value to 0.\x1b[0m"
                    )
                else:
                    total_tokens += log.get("total_tokens", 0)
                if log.get("total_cost") is None:
                    print(
                        f"\x1b[33mWarning: \"total_cost\" key is missing from log for file {file['name']}, defaulting value to 0.\x1b[0m"
                    )
                else:
                    total_cost += log.get("total_cost", 0.0)

        return log_files, total_tokens, total_cost, user_set

    def get_chat_history(self, branch_name: str, session_id: str) -> dict:
        """
        Retrieves chat history for a specific session.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.

        Returns:
            dict: The chat history for the session.
        """
        with open(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log", "r"
        ) as file:
            chat_history = yaml.safe_load(file)
        return chat_history

    def get_log_file_template(self, filename: str, branch: str) -> str:
        """
        Gets the Jinja template for a log file.

        Args:
            filename (str): The name of the log file.
            branch (str): The branch the log file is in.

        Returns:
            str: The rendered template for the log file.
        """

        def process_messages(messages: dict) -> dict:
            """
            Process each system response in the conversation and convert it to HTML.

            Args:
                filename (str): The name of the log file.

            Returns:
                dict: The processed messages in HTML format.
            """
            for message in messages:
                message["content"] = get_response_html(message["content"])

            return messages

        try:
            print(f"{self.chat_log_path}/chat_history/{branch}/{filename}")
            with open(
                f"{self.chat_log_path}/chat_history/{branch}/{filename}", "r"
            ) as file:
                content = yaml.safe_load(file)

            user_name = content["user"]
            real_name = content["real_name"]
            branch = content["branch"]
            time = content["time"]
            total_cost = round(content["total_cost"], 3)
            total_tokens = content["total_tokens"]

            try:
                messages = process_messages(content["messages"])
            except KeyError:
                messages = None

            return render_template(
                "display_chat_log.html",
                user_name=user_name,
                real_name=real_name,
                branch=branch,
                time=time,
                total_cost=total_cost,
                total_tokens=total_tokens,
                messages=messages,
                app_name=branch,
            )
        except FileNotFoundError:
            abort(404, description="Log file not found")
        except yaml.YAMLError as e:
            abort(500, description=f"Error parsing log file: {e}")

    def save_feedback(self, feedback: dict) -> None:
        """
        Saves feedback input to a YAML file.

        Args:
            feedback (dict): The feedback to save.

        Returns:
            None
        """

        # Make directory if it doesn't exist
        try:
            mkdir(f"{self.chat_log_path}/feedback")
        except FileExistsError:
            pass

        now = time.time()
        timestamp = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(now))
        filename = f"{self.chat_log_path}/feedback/{timestamp}.log"

        with open(filename, "w") as f:
            yaml.dump(feedback, f)

        print(f"Feedback saved to {filename}")

    def save_training_data(self, training_data: dict) -> None:
        """
        Saves training data to a YAML file.

        Args:
            training_data (dict): The training data to save.
        """

        # Make directory if it doesn't exist
        try:
            mkdir(f"{self.chat_log_path}/training_data")
        except FileExistsError:
            pass

        now = time.time()
        timestamp = time.strftime("%Y-%m-%d-%H-%M-%S", time.localtime(now))
        filename = f"{self.chat_log_path}/training_data/{timestamp}.log"

        with open(filename, "w") as f:
            yaml.dump(training_data, f)

        print(f"Training data saved to {filename}")

    def _get_file_list(self) -> list[dict]:
        """
        Get the list of chat history files with metadata.

        Returns:
            bool: True if the log file exists, False otherwise.
        """

        def get_creation_time(file_path):
            if platform.system() == "Darwin":  # macOS
                result = subprocess.run(
                    ["stat", "-f", "%B", file_path], capture_output=True, text=True
                )
                if result.returncode != 0:
                    raise RuntimeError(f"Error getting creation time: {result.stderr}")
                return int(result.stdout.strip())
            elif platform.system() == "Linux":
                result = subprocess.run(
                    ["stat", "-c", "%W", file_path], capture_output=True, text=True
                )
                if result.returncode != 0:
                    raise RuntimeError(f"Error getting creation time: {result.stderr}")
                return int(result.stdout.strip())
            else:
                # Fallback for other operating systems
                return int(path.getctime(file_path))

        def get_file_info(file_path: str) -> dict:
            """
            Get detailed information from a file and return it as a dictionary.

            Args:
                file_path (str): The path to the file.

            Returns:
                dict: A dictionary containing detailed information about the file.
            """

            def has_feedback(msgs: list) -> bool:
                for msg in msgs:
                    if "liked" in msg:
                        return True
                return False

            file_info = {}
            try:
                with open(file_path, "r") as file:
                    chat_log = yaml.safe_load(file)
                    file_info["has_feedback"] = has_feedback(
                        chat_log.get("messages", [])
                    )
                    file_info["first_message"] = (
                        chat_log.get("messages", [{}])[0]["content"]
                        if len(chat_log.get("messages", [])) > 0
                        else None
                    )
                    file_info["user"] = chat_log.get("user", "unknown user")
                    file_info["real_name"] = chat_log.get("real_name", "Student")
            except Exception as e:
                print(f"Error: Cannot read file {file_path}: {e}")
            return file_info

        file_list = []
        for root, dirs, files in walk(self.chat_log_path + "/chat_history"):
            for file_name in files:
                file_path = path.join(root, file_name)
                if path.isfile(file_path):  # Check if the path is a file
                    try:
                        created_time = get_creation_time(file_path)
                    except RuntimeError:
                        # Fallback if stat doesn't work at all (may show modified time)
                        created_time = int(path.getctime(file_path))

                    file_stat = stat(file_path)
                    file_info = {
                        "name": file_name,
                        "created": created_time,
                        "modified": file_stat.st_mtime,
                        "branch": path.basename(
                            root
                        ),  # Get the branch name from the directory
                    }
                    # Update file_info with additional details from get_file_info
                    file_info.update(get_file_info(file_path))
                    file_list.append(file_info)
        return file_list

    def _create_log_file(
        self, branch_name: str, session_id: str, user: User | None = None
    ) -> None:
        """
        Creates a new log file for a chat session.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            user (User | None): Optional User to obtain information from to include in the log.

        Returns:
            None
        """
        # compile log information
        log_info: dict = {
            "session_id": session_id,
            "user": user.full_id_name if user else "anon",
            "real_name": user.realname if user else "anon",
            "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "branch": branch_name,
            "total_cost": 0,
            "total_tokens": 0,
            "messages": [],
        }

        # ensure log directory exists
        if not path.exists(f"{self.chat_log_path}/chat_history/{branch_name}"):
            makedirs(f"{self.chat_log_path}/chat_history/{branch_name}")

        # create log file
        with open(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log", "w"
        ) as file:
            yaml.dump(log_info, file)

    def _update_log_file(
        self, branch_name: str, session_id: str, log_data: dict
    ) -> None:
        """
        Updates the log file with the new log data.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            log_data (dict): The data to be logged. Should contain the following keys: "user_info", "cost", "tokens_used", and "message".

        Returns:
            None
        """
        with open(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log", "r"
        ) as file:
            log: dict = yaml.safe_load(file)

            log["messages"] = log.get("messages", [])

            # Add user message
            log["messages"].append(
                {
                    "role": "user",
                    "content": log_data["messages"][-2],
                }
            )

            # Add chatbot message and execution stats
            log["messages"].append(
                {
                    "role": "system",
                    "content": log_data["messages"][-1],
                    "context": [
                        context.page_content
                        for context in log_data["retrieved_context"]
                    ],
                    "execution_time": log_data.get("execution_time", 0),
                    "tokens_used": log_data.get("tokens_used", 0),
                    "cost": log_data.get("cost", 0),
                }
            )

            log["total_cost"] += log_data.get("cost", 0)
            log["total_tokens"] += log_data.get("tokens_used", 0)

        with open(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log", "w"
        ) as file:
            yaml.dump(log, file)

    def _does_log_exist(self, branch_name: str, session_id: str) -> bool:
        """
        Checks if a log file exists for the given session ID.

        Args:
            session_id (str): The session ID to check for.

        Returns:
            bool: True if the log file exists, False otherwise.
        """
        return path.exists(
            f"{self.chat_log_path}/chat_history/{branch_name}/{session_id}.log"
        )



"""Module for handling chat interface rendering.

This module contains a function to render the chat interface template with relevant data.

© 2024 Blaine Freestone, Carson Bush

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import render_template

from maeser.chat.chat_logs import BaseChatLogsManager
# from flask_login import current_user

from maeser.chat.chat_session_manager import ChatSessionManager

def controller(
        chat_sessions_manager: ChatSessionManager,
        max_requests: int | None = None,
        rate_limit_interval: int | None = None,
        current_user=None,
        app_name: str | None = None,
        main_logo_login: str | None = None,
        main_logo_chat: str | None = None,
        chat_head: str | None = None,
        favicon: str | None = None,        
    ):
    """
    Renders the chat interface template with relevant data.

    Args:
        chat_sessions_manager (ChatSessionManager): The chat session manager object.
        max_requests (int, optional): The maximum number of requests a user can make. Defaults to None.
        rate_limit_interval (int, optional): The interval in seconds for rate limiting requests. Defaults to None.
        current_user (object, optional): The current user object. Defaults to None.

    Returns:
        The rendered 'chat_interface.html' template with the following data:
            - conversation: None (no active conversation)
            - buttons: The dictionary of available chat branches
            - links: A list of dictionaries representing previous chat sessions for the current user
            - requests_remaining: The number of requests remaining for the current user (10 if current_user is None)
            - max_requests_remaining: The maximum number of requests allowed
            - requests_remaining_interval_ms: The interval in milliseconds for rate limiting requests (rate_limit_interval * 1000 / 3)
    """
    # Get chat log path and branches from chat sessions
    log_manager: BaseChatLogsManager | None = chat_sessions_manager.chat_logs_manager
    chat_branches: dict = chat_sessions_manager.branches

    links = log_manager.get_chat_history_overview(current_user) if log_manager else []

    # Get rate limiting information if enabled
    requests_remaining: int | None = None if current_user is None else current_user.requests_remaining
    if rate_limit_interval:
        rate_limit_interval = rate_limit_interval * 1000 // 3
    rate_limiting = bool(requests_remaining and rate_limit_interval and max_requests)

    user_management = True if current_user else False

    is_admin = current_user.admin if current_user else False
    
    return render_template(
        'chat_interface.html',
        conversation=None,
        buttons=chat_branches,                                  # dict
        links=links,                                            # List[dict]
        requests_remaining=requests_remaining,                  # None | int
        max_requests_remaining=max_requests,                    # None | int
        requests_remaining_interval_ms=rate_limit_interval,     # None | int
        rate_limiting=rate_limiting,                            # str
        user_management=user_management,                        # str
        main_logo_login=main_logo_login,                        # None | str
        main_logo_chat=main_logo_chat,                          # None | str
        chat_head=chat_head,                                    # None | str
        favicon=favicon,                                        # None | str
        app_name=app_name if app_name else "Maeser",            # str
        is_admin=is_admin,                                      # bool
        str=str,
    )



"""
Module for handling conversation history retrieval in a Flask API.

This module defines a controller function that retrieves the conversation history
for a given session and branch. The conversation history is processed to handle
system messages by applying HTML response formatting.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import jsonify, request
from maeser.render import get_response_html

def controller(session_handler):
    """
    Retrieves the conversation history for a given session and branch.

    Args:
        session_handler (object): An object that handles session management
            and provides a method to retrieve conversation history.

    Returns:
        dict: A dictionary containing the conversation history, with system
            messages having their content processed by the `get_response_html` function.

    The function expects the request data to contain 'session' and 'branch' keys,
    which are used to retrieve the conversation history from the `session_handler` object.
    If the conversation history contains 'messages', it iterates through them and processes
    the content of system messages using the `get_response_html` function. Finally, it
    returns the conversation history as a JSON response.
    """
    data = request.get_json()

    session = data.get('session')
    branch = data.get('branch')
    
    conversation_history = session_handler.get_conversation_history(branch, session)
    if 'messages' in conversation_history:
        for message in conversation_history['messages']:
            if message['role'] == 'system':
                message['content'] = get_response_html(message['content'])
            else:
                message['content'] = message['content']

    return jsonify(conversation_history)



"""
Markdown response conversion module. Intended for use with LLM output.

This module provides a utility function to easily customize HTML layouts and convert markdown 
responses to HTML with additional processing, such as adding target="_blank" to anchor tags 
and adjusting paths for images.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

import markdown

def get_response_html(response: str) -> str:
    """
    Convert a markdown response to HTML.

    Args:
        response (str): The markdown response.

    Returns:
        str: The HTML response.
    """
    text = response
    html_content = markdown.markdown(text, extensions=['pymdownx.superfences', 'tables', 'smarty', 'sane_lists'])
    # Add target="_blank" attribute to anchor tags
    html_content = html_content.replace('<a href', '<a target="_blank" href')
    html_content = html_content.replace('figures/', '/figures/')
    return html_content


"""Module for handling feedback for messages through a controller.

This module defines a controller function that processes feedback
for messages. The feedback includes information such as the branch,
session ID, message, whether it is a like or dislike, and the index
of the message.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager
from flask import request

def controller(session_handler: ChatSessionManager):
    """
    Handle feedback for messages.

    Args:
        session_handler (ChatSessionManager): The session handler to
            manage chat sessions and feedback.

    Returns:
        dict: Status of the feedback submission.
    """
    data = request.get_json()
    branch = data.get('branch')
    session_id = data.get('session_id')
    message = data.get('message')
    like = data.get('like')
    index = int(data.get('index'))
    print(f'Received feedback: {"Like" if like else "Dislike"} for message: {message} at index: {index}')
    # Handle the feedback (e.g., save to a database, log it, etc.)
    session_handler.add_feedback(branch, session_id, index, like)
    return {'status': 'success'}



"""
This is the controllers subpackage for the Maeser package. It contains the
controllers for the different parts of the application.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from . import (
    chat_api,
    chat_interface,
    chat_logs_overview,
    display_chat_log,
    feedback_api,
    feedback_form_get,
    feedback_form_post,
    login_api,
    logout,
    new_session_api,
    training,
    training_post,
    conversation_history_api,
)
from . import common

__all__ = [
    'chat_api',
    'chat_interface',
    'chat_logs_overview',
    'display_chat_log',
    'feedback_api',
    'feedback_form_get',
    'feedback_form_post',
    'login_api',
    'logout',
    'new_session_api',
    'training',
    'training_post',
    'conversation_history_api',
    'common',
]


"""
This module handles API requests for creating new chat sessions.

It uses the ChatSessionManager to manage session creation and optionally integrates
with user management through Flask-Login's current_user.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager
from flask import request
from flask_login import current_user

def controller(session_handler: ChatSessionManager, user_management: bool = False):
    """
    Handle session requests.

    Args:
        session_handler (ChatSessionManager): The session handler instance.
        user_management (bool): Flag to indicate if user management is enabled.

    Returns:
        dict: Response confirming session action or an error message.
    """
    posty = request.get_json()
    branch_action = posty['action']
    if posty['type'] == 'new':
        if user_management:
            return {'response': session_handler.get_new_session_id(branch_action, current_user)} # type: ignore
        return {'response': session_handler.get_new_session_id(branch_action)}
    return {'response': 'invalid', 'details': 'Requested session type is not valid'}



"""
User management module for authentication and authorization.

This module provides classes and utilities for managing users,
including authentication methods, database operations, and request tracking.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""


import secrets
import sqlite3
from abc import ABC, abstractmethod
from typing import Any, Tuple, Union
from urllib.parse import urlencode
import os
import ssl
from ldap3 import Server, Connection, ALL, SUBTREE, Tls
from ldap3.core.exceptions import LDAPException, LDAPAttributeError, LDAPBindError, LDAPSocketReceiveError

import requests


class User:
    """
    Provides default implementations for the methods that Flask-Login expects user objects to have.
    """

    # Python 3 implicitly sets __hash__ to None if we override __eq__
    # We set it back to its default implementation
    __hash__ = object.__hash__

    def __init__(self, ident: str, blacklisted=False, admin=False, realname='Student', usergroup='b\'guest\'', authmethod='invalid', requests_left=10, max_requests=10, aka=[]):
        """
        Initialize a User object.

        Args:
            ident (str): The user's identifier.
            blacklisted (bool, optional): Whether the user is blacklisted. Defaults to False.
            admin (bool, optional): Whether the user is an admin. Defaults to False.
            realname (str, optional): The user's real name. Defaults to 'Student'.
            usergroup (str, optional): The user's group. Defaults to 'b\'guest\''.
            authmethod (str, optional): The authentication method. Defaults to 'invalid'.
            requests_left (int, optional): The number of requests left. Defaults to 10.
            max_requests (int, optional): The maximum number of requests. Defaults to 10.
            aka (list, optional): A list of alternate names. Defaults to an empty list.
        """
        self.ident = ident
        self.is_active = not blacklisted
        self.admin = admin
        self.realname = realname
        self.usergroup = usergroup
        self.auth_method = authmethod
        self._requests_remaining = requests_left
        self._max_requests = max_requests
        self.aka: list = aka
    
    def __str__(self) -> str:
        return f"""User Information for {self.ident}:
        Authentication Method: {self.auth_method}
        Real Name: {self.realname}
        Admin: {'Yes' if self.admin else 'No'}
        Banned: {'Yes' if not self.is_active else 'No'}
        User Group: {self.usergroup}
        Requests Remaining: {self.requests_remaining}/{self._max_requests}"""
        
    @property
    def json(self) -> dict[str, Any]:
        return {
            'ident': self.ident,
            'is_active': self.is_active,
            'admin': self.admin,
            'realname': self.realname,
            'usergroup': self.usergroup,
            'auth_method': self.auth_method,
            'requests_remaining': self.requests_remaining,
            'max_requests': self._max_requests,
            'aka': self.aka
        }

    @property
    def is_authenticated(self):
        """Return True if the user is authenticated."""
        return self.is_active

    @property
    def is_anonymous(self):
        """Return False, as anonymous users are not supported."""
        return False

    def get_id(self):
        """Return the user's full identifier name including authentication method."""
        return self.full_id_name

    @property
    def full_id_name(self):
        """Return the user's full identifier name including authentication method."""
        return f'{self.auth_method}.{self.ident}'

    @property
    def requests_remaining(self):
        """Return the number of requests remaining for the user."""
        return self._requests_remaining

    @requests_remaining.setter
    def requests_remaining(self, num: int):
        """
        Set the number of requests remaining for the user.

        Args:
            num (int): The new number of requests remaining.
        """
        if num >= self._max_requests:
            self._requests_remaining = self._max_requests
        elif num <= 0:
            self._requests_remaining = 0
        else:
            self._requests_remaining = num

    def __eq__(self, other):
        """
        Check the equality of two User objects using get_id.

        Args:
            other (User): The other user to compare.

        Returns:
            bool: True if the users are equal, False otherwise.
        """
        if isinstance(other, User):
            return self.get_id() == other.get_id()
        return NotImplemented

    def __ne__(self, other):
        """
        Check the inequality of two User objects using get_id.

        Args:
            other (User): The other user to compare.

        Returns:
            bool: True if the users are not equal, False otherwise.
        """
        equal = self.__eq__(other)
        return not equal


class LoginStyle:
    
    def __init__(self, icon: str, login_submit: str, direct_submit: bool=False):
        # Not a url, but a controller name for url_for. i.e. 'maeser.github_authorize' or 'localauth'
        self.login_submit = login_submit
        self.direct_submit = direct_submit
        # HTML for a custom form (labels and inputs only)
        self._custom_form: str = '<label for="username" class="form-label">Username</label><input type="text" id="username" name="username" class="form-input" required><label for="password" class="form-label">Password</label><input type="password" id="password" name="password" class="form-input" required>'
        self.icon_html = f'<i class="bi bi-{icon}"></i>'

    @property
    def form_html(self) -> str:
        if self.direct_submit:
            raise ValueError("Cannot use form_html with direct_submit=True")
        return self._custom_form
    
    @form_html.setter
    def form_html(self, html: str):
        self._custom_form = html


class BaseAuthenticator(ABC):
    """
    Base class for authenticators.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        """
        Initialize the authenticator with any required arguments.
        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Return the string representation of the authenticator."""
        pass

    @abstractmethod
    def authenticate(self, *args, **kwargs) -> Union[tuple, None]:
        """
        Authenticate a user.

        Args:
            *args: Positional arguments for authentication.
            **kwargs: Keyword arguments for authentication.

        Returns:
            tuple or None: A tuple containing the user's username, real name, and user group if authentication is successful, otherwise None.
        """
        pass

    @abstractmethod
    def fetch_user(self, ident: str) -> Union[User, None]:
        """
        Fetch a user from the authenticator.

        Args:
            ident (str): The identifier of the user to fetch.

        Returns:
            User or None: The fetched user object or None if not found.
                ENSURE THAT YOU SET max_requests TO THE CORRECT VALUE FOR THE USER!
        """
        pass
    
    @property
    @abstractmethod
    def style(self) -> LoginStyle:
        """
        Get the login style for the authenticator.

        Returns:
            LoginStyle: The login style object.
        """
        pass


class GithubAuthenticator(BaseAuthenticator):
    """
    Handles authentication with GitHub OAuth.
    """

    def __init__(self, client_id: str, client_secret: str, auth_callback_uri: str, timeout: int = 10, max_requests: int = 10):
        """
        Initialize the GitHub authenticator.

        Args:
            client_id (str): The GitHub client ID.
            client_secret (str): The GitHub client secret.
            auth_callback_uri (str): The callback URI for GitHub authentication.
        """
        self.client_id = client_id
        self.client_secret = client_secret
        # Generally this should be set from your Flask app as this will differ between applications
        # url_for('github_auth_callback', _external=True)
        self._max_requests = max_requests
        self.auth_callback_uri = auth_callback_uri
        self.timeout = timeout
        self._login_style = LoginStyle('github', 'maeser.github_authorize', direct_submit=True)

    def __str__(self) -> str:
        return 'GitHub'
    
    @property
    def style(self) -> LoginStyle:
        return self._login_style

    def authenticate(self, request_args: dict, oauth_state: str) -> Union[tuple, None]:
        """
        Authenticate a user with GitHub OAuth.

        Args:
            request_args (dict): The request arguments containing the authorization code and state.
            oauth_state (str): The state value used to prevent CSRF attacks.

        Returns:
            tuple or None: A tuple containing the user's username, real name, and user group if authentication is successful, otherwise None.
        """
        if request_args['state'] != oauth_state or 'code' not in request_args:
            print(request_args['state'], oauth_state, 'ERROR') 
            return None

        token_url = 'https://github.com/login/oauth/access_token'
        user_info_url = 'https://api.github.com/user'

        # exchange the authorization code for an access token
        response = requests.post(token_url, data={
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'code': request_args['code'],
            'grant_type': 'authorization_code',
            'redirect_uri': self.auth_callback_uri
        }, headers={'Accept': 'application/json'},
        timeout=self.timeout)

        if response.status_code != 200:
            print(f'GitHub authentication failed during token exchange: {response.status_code}', 'ERROR')
            return None

        oauth2_token = response.json().get('access_token')
        if not oauth2_token:
            print('GitHub authentication failed: No access token received', 'ERROR')
            return None

        response = requests.get(user_info_url, headers={
            'Authorization': 'Bearer ' + oauth2_token,
            'Accept': 'application/json',
        }, timeout=self.timeout)

        if response.status_code != 200:
            print(f'GitHub authentication failed when fetching user info: {response.status_code}', 'ERROR')
            return None

        json_response = response.json()
        print(json_response)
        return json_response['login'], json_response['name'], 'b\'guest\''

    def fetch_user(self, ident: str) -> Union[User, None]:
        """
        Fetch a user from the GitHub API.

        Args:
            ident (str): The username of the user to fetch.

        Returns:
            User or None: The fetched user object or None if the user is not found.
        """
        user_info_url = f'https://api.github.com/users/{ident}'
        response = requests.get(user_info_url)
        if response.status_code == 200:
            json_response = response.json()
            return User(json_response['login'], realname=json_response.get('name', ''), usergroup='b\'guest\'', authmethod='github', max_requests=self._max_requests)
        print(f'WARNING: No GitHub user "{ident}" found')
        return None

    def get_auth_info(self) -> Tuple[str, str]:
        """
        Get the GitHub authorization information.

        Returns:
            tuple: A tuple containing the OAuth state and provider URL.
        """
        authorize_url = 'https://github.com/login/oauth/authorize'
        scopes = ['user:email']

        # generate a random string for the state parameter
        oauth_state = secrets.token_urlsafe(16)

        query_string = urlencode({
            'client_id': self.client_id,
            'redirect_uri': self.auth_callback_uri,
            'response_type': 'code',
            'scope': ' '.join(scopes),
            'state': oauth_state,
        })

        provider_url = authorize_url + '?' + query_string

        return oauth_state, provider_url


class LDAPAuthenticator(BaseAuthenticator):
    def __init__(self, 
                 name: str,
                 ldap_server_urls: list, 
                 ldap_base_dn: str, 
                 attribute_name: str, 
                 search_filter: str, 
                 object_class: str, 
                 attributes: list, 
                 ca_cert_path: str = '/etc/ssl/certs', 
                 connection_timeout: int = 5):
        
        self.name = name
        self.ldap_server_urls = ldap_server_urls
        self.ldap_base_dn = ldap_base_dn
        self.attribute_name = attribute_name
        self.search_filter = search_filter
        self.object_class = object_class
        self.attributes = attributes
        self.ca_cert_path = ca_cert_path
        self.connection_timeout = connection_timeout

        # Ensure certificate directory exists
        if not os.path.exists(self.ca_cert_path):
            raise FileNotFoundError('Path to CA Certificates directory does not exist')

        # Initialize LDAP server instances
        self.ldap_servers = self._initialize_ldap_servers()
        self.ldap_usable_servers = self._test_ldap_anonymous_bind()
        self._next_server_index = 0
        self._login_style = LoginStyle('envelope-fill', 'maeser.login', direct_submit=False)

    def __str__(self):
        return self.name

    @property
    def style(self):
        return self._login_style
    
    @property
    def next_ldap_server(self)-> Union[Server, None]:
        """Return the next available LDAP server in a round-robin fashion."""
        if len(self.ldap_usable_servers) == 0:
            print("NO REACHABLE LDAP SERVER!")
            return None
        server_to_use = self.ldap_usable_servers[self._next_server_index]
        self._next_server_index = (self._next_server_index + 1) % len(self.ldap_usable_servers)
        return server_to_use

    def _initialize_ldap_servers(self) -> list[Server]:
        """
        Initialize LDAP server instances with retrieved certificates.

        Returns:
            list: A list of initialized LDAP Server objects.
        """
        servers: list[Server] = []
        for server_url in self.ldap_server_urls:
            try:
                servers.append(Server(
                    server_url,
                    use_ssl=True,
                    get_info=ALL,
                    connect_timeout=self.connection_timeout,
                    tls=Tls(validate=ssl.CERT_REQUIRED, ca_certs_path=self.ca_cert_path)
                ))
            except LDAPException as e:
                print(f'Unable to initialize LDAP server {server_url}: {type(e)}, {e}')
        return servers

    def _test_ldap_anonymous_bind(self) -> list:
        """
        Test anonymous bind to each LDAP server and blacklist bad servers.

        Returns:
            list: A list of LDAP servers that are usable.
        """
        usable_servers = []
        for ldap_server in self.ldap_servers:
            try:
                test_connection = Connection(ldap_server, receive_timeout=self.connection_timeout)
                if test_connection.bind():
                    usable_servers.append(ldap_server)
                test_connection.unbind()
            except (LDAPException, LDAPAttributeError, LDAPBindError, LDAPSocketReceiveError) as e:
                print(f'Failed to bind to LDAP server {ldap_server}: {type(e)}, {e}')
        return usable_servers

    def authenticate(self, ident: str, password: str) -> Union[tuple, None]:
        if self.next_ldap_server is None:
            return None
        try:
            conn = Connection(
                self.next_ldap_server,
                user=f'{self.attribute_name}={ident},{self.ldap_base_dn}',
                password=password,
                auto_bind=True,
                read_only=True,
                receive_timeout=self.connection_timeout
            )
        except (LDAPException, LDAPAttributeError, LDAPBindError, LDAPSocketReceiveError) as e:
            print(f'{self.name} user {ident} failed to authenticate: {type(e)}: {e}')
            return None

        try:
            conn.search(
                self.ldap_base_dn,
                self.search_filter.format(ident=ident),
                SUBTREE,
                attributes=self.attributes,
                time_limit=int(self.connection_timeout)
            )
            
            if conn.entries:
                display_name = conn.entries[0].displayName.value
                user_group = conn.entries[0].memberOf.value if 'memberOf' in self.attributes else None
                return ident, display_name, user_group
            
        except LDAPSocketReceiveError as e:
            print(f'LDAP search timed out for user {ident}: {e}')
            
        finally:
            conn.unbind()
        
        return None

    def fetch_user(self, ident: str) -> Union[User, None]:
        """
        Fetch user information from LDAP and return a User object.
        This method performs an anonymous bind and searches for the user.
        No authentication is preformed using this method.

        Args:
            ident (str): The user's identifier.

        Returns:
            Union[User, None]: The User object if found, None otherwise.
        """
        if self.next_ldap_server is None:
            return None
        try:
            conn = Connection(self.next_ldap_server, auto_bind=True, receive_timeout=self.connection_timeout)
        except (LDAPException, LDAPAttributeError, LDAPBindError, LDAPSocketReceiveError) as e:
            print(f'LDAP fetch for user {ident} failed: {type(e)}, {e}')
            return None
        
        try:
            search_filter = self.search_filter.format(ident=ident)
            conn.search(
                self.ldap_base_dn, 
                search_filter, 
                SUBTREE, 
                attributes=self.attributes,
                time_limit=int(self.connection_timeout)
            )
            
            if conn.entries:
                display_name = conn.entries[0].displayName.value
                user_group = conn.entries[0].memberOf.value if 'memberOf' in self.attributes else None
                return User(ident, realname=display_name, usergroup=user_group, authmethod=self.name)
        except LDAPSocketReceiveError as e:
            print(f'LDAP search timed out for user {ident}: {e}')
        finally:
            conn.unbind()

        print(f'No {self.name} user "{ident}" found')
        return None


class UserManager:
    """
    Manages user operations including authentication, database interactions, and request tracking.
    """

    def __init__(self, db_file_path: str, max_requests: int = 10, rate_limit_interval: int = 180):
        """
        Initialize the UserManager.

        Args:
            db_file_path (str): The file path to the SQLite database.
            max_requests (int, optional): The maximum number of requests a user can have. Defaults to 10.
        """
        self.db_file_path = db_file_path
        self.authenticators: dict[str, BaseAuthenticator] = {}
        self.max_requests = max_requests
        self.rate_limit_interval = rate_limit_interval
        self._create_tables()

    def register_authenticator(self, name: str, authenticator: BaseAuthenticator):
        """
        Register a new authentication method.

        Args:
            name (str): The shorthand name of the authentication method. Must only contain letters.
            authenticator (BaseAuthenticator): The authenticator object.

        Raises:
            ValueError: If the provided name is invalid or the authenticator is already registered.
        """
        if not name.isalpha():
            raise ValueError(f"Invalid authenticator name: {name}, must only contain letters!")
        self.authenticators[name] = authenticator
        with self.db_connection as db:
            self._create_table(db, name)

    @property
    def db_connection(self) -> sqlite3.Connection:
        """
        Open a connection to the SQLite database.

        Returns:
            sqlite3.Connection: The database connection.

        Raises:
            sqlite3.OperationalError: If the database cannot be opened.
        """
        try:
            return sqlite3.connect(self.db_file_path)
        except sqlite3.OperationalError as e:
            print(
                f"\033[31mUnable to open sqlite db file {self.db_file_path}, using tempory storage: {e}\033[0m"
            )
            return sqlite3.connect(":memory:")

    def _create_tables(self):
        with self.db_connection as db:
            for auth_method in self.authenticators:
                self._create_table(db, auth_method)

    def _create_table(self, db: sqlite3.Connection, auth_method: str):
        if not auth_method.isalnum():
            raise ValueError(f"Invalid authenticator name: {auth_method}")

        table_name = f"{auth_method}Users"
        db.execute(f'''
            CREATE TABLE IF NOT EXISTS "{table_name}" (
                user_id TEXT PRIMARY KEY,
                blacklisted BOOL,
                admin BOOL,
                realname TEXT,
                usertype TEXT,
                requests_left INT,
                aka TEXT
            )
        ''')

    def check_user_auth(self, auth_method: str) -> bool:
        return auth_method in self.authenticators

    def get_user(self, auth_method: str, ident: str) -> Union[User, None]:
        """
        Retrieve a user from the database.

        Args:
            auth_method (str): The authentication method used.
            ident (str): The unique identifier of the user.

        Returns:
            User: The user object, or None if not found.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if not auth_method.isalnum():
            raise ValueError(f"Invalid authenticator name: {auth_method}")

        table_name = f"{auth_method}Users"
        with self.db_connection as db:
            cursor: sqlite3.Cursor = db.execute(
                f'SELECT user_id, blacklisted, admin, realname, usertype, requests_left FROM "{table_name}" WHERE user_id=?',
                (ident,)
            )
            row = cursor.fetchone()
            if row:
                return User(row[0], bool(row[1]), bool(row[2]), realname=row[3], usergroup=str(row[4]), requests_left=row[5], authmethod=auth_method, max_requests=self.max_requests)
        return None

    def list_users(self, auth_filter: str | None = None, admin_filter: str | None = None, banned_filter: str | None = None) -> list[User]:
        """
        List all users in the database, optionally filtered by authentication method, admin status, and banned status.

        Args:
            auth_filter (str, optional): The authentication method to list users for. If None or 'all', list users from all authentication methods.
            admin_filter (str, optional): Filter users by admin status. Can be 'all', 'admin', or 'non-admin'.
            banned_filter (str, optional): Filter users by banned status. Can be 'all', 'banned', or 'non-banned'.

        Returns:
            list[User]: A list of user objects.

        Raises:
            ValueError: If the provided auth_method is invalid or if admin_filter or banned_filter have invalid values.
        """
        if auth_filter is not None and auth_filter != 'all' and not auth_filter.isalnum():
            raise ValueError(f"Invalid authenticator name: {auth_filter}")

        if admin_filter is not None and admin_filter not in ['all', 'admin', 'non-admin']:
            raise ValueError(f"Invalid admin_filter value: {admin_filter}")

        if banned_filter is not None and banned_filter not in ['all', 'banned', 'non-banned']:
            raise ValueError(f"Invalid banned_filter value: {banned_filter}")

        users: list[User] = []
        with self.db_connection as db:
            if auth_filter is None or auth_filter == 'all':
                cursor: sqlite3.Cursor = db.execute("SELECT name FROM sqlite_master WHERE type='table' AND name LIKE '%Users'")
                tables = [table_name[0] for table_name in cursor.fetchall()]
            else:
                tables = [f"{auth_filter}Users"]

            for table_name in tables:
                query = f'SELECT user_id, blacklisted, admin, realname, usertype, requests_left FROM "{table_name}"'
                conditions = []
                
                if admin_filter == 'admin':
                    conditions.append("admin = 1")
                elif admin_filter == 'non-admin':
                    conditions.append("admin = 0")

                if banned_filter == 'banned':
                    conditions.append("blacklisted = 1")
                elif banned_filter == 'non-banned':
                    conditions.append("blacklisted = 0")
                
                if conditions:
                    query += " WHERE " + " AND ".join(conditions)

                cursor = db.execute(query)
                for row in cursor.fetchall():
                    auth_method_from_table = table_name.replace("Users", "")
                    users.append(User(row[0], bool(row[1]), bool(row[2]), realname=row[3], usergroup=str(row[4]), requests_left=row[5], authmethod=auth_method_from_table, max_requests=self.max_requests))

        return users

    def authenticate(self, auth_method: str, *args: Any, **kwargs: Any) -> Union[User, None]:
        """
        Authenticate a user using the specified authentication method.

        Args:
            auth_method (str): The authentication method to use.
            *args: Positional arguments for the authentication method.
            **kwargs: Keyword arguments for the authentication method.

        Returns:
            User: The authenticated user object, or None if authentication fails.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        authenticator = self.authenticators.get(auth_method)
        if not authenticator:
            raise ValueError(f"Unsupported authentication method: {auth_method}")

        auth_result = authenticator.authenticate(*args, **kwargs)
        print(auth_result)
        if auth_result:
            user_id, display_name, user_group = auth_result
            return self._create_or_update_user(auth_method, user_id, display_name, user_group)
        return None

    def _create_or_update_user(self, auth_method: str, user_id: str, display_name: str, user_group: str) -> User:
        """
        Create or update a user in the database.

        Args:
            auth_method (str): The authentication method used.
            user_id (str): The unique identifier of the user.
            display_name (str): The display name of the user.
            user_group (str): The group the user belongs to.

        Returns:
            User: The user object.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Unsupported authentication method: {auth_method}")
        with self.db_connection as db:
            table_name = f"{auth_method}Users"
            cursor = db.execute(f'SELECT user_id, blacklisted, admin, realname, usertype, requests_left FROM "{table_name}" WHERE user_id=?', (user_id,))
            row = cursor.fetchone()

            if row:
                user = User(row[0], bool(row[1]), bool(row[2]), realname=row[3], requests_left=row[5], authmethod=auth_method, max_requests=self.max_requests)
            else:
                db.execute(
                    f'INSERT INTO "{table_name}" (user_id, blacklisted, admin, realname, usertype, requests_left) VALUES (?, ?, ?, ?, ?, ?)',
                    (str(user_id), False, False, str(display_name), str(user_group), int(self.max_requests))
                )
                db.commit()
                user = User(user_id, realname=display_name, usergroup=user_group, authmethod=auth_method, max_requests=self.max_requests)

        return user

    def update_admin_status(self, auth_method: str, ident: str, is_admin: bool):
        """
        Update the admin status of a user.

        Args:
            auth_method (str): The authentication method used.
            ident (str): The identifier of the user.
            is_admin (bool): Whether the user should be an admin or not.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")

        table_name = f"{auth_method}Users"
        with self.db_connection as db:
            db.execute(f'UPDATE "{table_name}" SET admin=? WHERE user_id=?', (is_admin, ident))
            db.commit()

    def update_banned_status(self, auth_method: str, ident: str, is_banned: bool):
        """
        Update the banned status of a user.

        Args:
            auth_method (str): The authentication method used.
            ident (str): The identifier of the user.
            is_banned (bool): Whether the user should be banned or not.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")

        table_name = f"{auth_method}Users"
        with self.db_connection as db:
            db.execute(f'UPDATE "{table_name}" SET blacklisted=? WHERE user_id=?', (is_banned, ident))
            db.commit()

    def refresh_requests(self, inc_by: int = 1):
        """
        Refresh the number of requests for all users by the given amount.

        Args:
            inc_by (int, optional): The amount to increase the requests by. Defaults to 1.
        """
        with self.db_connection as db:
            for auth_method in self.authenticators:
                table_name = f"{auth_method}Users"
                db.execute(f'''
                    UPDATE "{table_name}"
                    SET requests_left = MIN(?, MAX(0, requests_left + ?))
                ''', (self.max_requests, inc_by))
            db.commit()

    def decrease_requests(self, auth_method: str, user_id: str, dec_by: int = 1):
        """
        Decrease the number of requests remaining for a user.

        Args:
            auth_method (str): The authentication method used.
            user_id (str): The identifier of the user.
            dec_by (int, optional): The amount to decrease the requests by. Defaults to 1.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")
        
        dec_by = min(dec_by, self.max_requests)

        table_name = f"{auth_method}Users"
        with self.db_connection as db:
            db.execute(f'''
                UPDATE "{table_name}"
                SET requests_left = MAX(0, requests_left - ?)
                WHERE user_id = ?
            ''', (dec_by, user_id))
            db.commit()

    def increase_requests(self, auth_method: str, user_id: str, inc_by: int = 1):
        """
        Increase the number of requests remaining for a user.

        Args:
            auth_method (str): The authentication method used.
            user_id (str): The identifier of the user.
            inc_by (int, optional): The amount to increase the requests by. Defaults to 1.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")
        
        inc_by = min(inc_by, self.max_requests)

        table_name = f"{auth_method}Users"
        with self.db_connection as db:
            db.execute(f'''
                UPDATE "{table_name}"
                SET requests_left = MIN(?, MAX(0, requests_left + ?))
                WHERE user_id = ?
            ''', (self.max_requests, inc_by, user_id))

    def get_requests_remaining(self, auth_method: str, user_id: str) -> Union[int, None]:
        """
        Get the number of requests remaining for a user.

        Args:
            auth_method (str): The authentication method used.
            user_id (str): The identifier of the user.

        Returns:
            Union[int, None]: The number of requests remaining, or None if the user is not found.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")

        user = self.get_user(auth_method, user_id)
        return user.requests_remaining if user else None
    
    def fetch_user(self, auth_method: str, ident: str) -> bool:
        """
        Fetch a user from the authentication source and add them to the cache
        without modifying their admin or banned status.

        Args:
            auth_method (str): The authentication method.
            ident (str): The user's identifier.

        Returns:
            bool: True if the user was successfully fetched and cached, False otherwise.
        """
        if auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")
        
        user = self.authenticators[auth_method].fetch_user(ident)
        if user:
            self._create_or_update_user(auth_method, user.ident, user.realname, user.usergroup)
            return True
        return False
        
    def remove_user_from_cache(self, auth_method: str, ident: str, force_remove: bool = False) -> bool:
        """
        Remove a user from the cache.

        Args:
            auth_method (str): The authentication method used.
            ident (str): The identifier of the user.

        Returns:
            bool: True if the user was removed, False otherwise.

        Raises:
            ValueError: If the provided auth_method is invalid.
        """
        if not force_remove and auth_method not in self.authenticators:
            raise ValueError(f"Invalid authenticator name: {auth_method}")

        table_name = f"{auth_method}Users"
        with self.db_connection as db:
            cursor = db.execute(f'DELETE FROM "{table_name}" WHERE user_id=?', (ident, ))
            db.commit()
            return bool(cursor.rowcount)
        
    def list_cleanables(self):
        """
        List non-banned and non-admin users in the cache/database.
        
        Returns:
            list[str]: A list of user identifiers in the format "auth_method:user_id".
        """
        cleanables = []
        with self.db_connection as db:
            for auth_method in self.authenticators:
                table_name = f"{auth_method}Users"
                cursor = db.execute(f'SELECT user_id FROM "{table_name}" WHERE blacklisted=0 AND admin=0')
                cleanables.extend([f'{auth_method}:{row[0]}' for row in cursor.fetchall()])
        return cleanables

    def clean_cache(self) -> int:
        """
        Clean the cache by removing non-banned and non-admin users.

        Returns:
            int: The number of users removed from the cache.
        """
        removed_count = 0
        with self.db_connection as db:
            for auth_method in self.authenticators:
                table_name = f"{auth_method}Users"
                removed = db.execute(f'DELETE FROM "{table_name}" WHERE blacklisted=0 AND admin=0').rowcount
                removed_count += removed
        return removed_count



"""
User management API controller.

This module provides an API endpoint for managing users, including listing users,
toggling admin and banned statuses, updating request counts, removing users from
the cache, fetching users, listing cleanable users, and cleaning the cache.

© 2024 Carson Bush

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any
from maeser.user_manager import UserManager
from flask import abort, request

def controller(user_manager: UserManager) -> list[dict[str, Any]] | dict[str, str] | list[Any]:
    """
    API endpoint for user management.

    Args:
        user_manager (UserManager): User Manager object to interact with the user database.

    Returns:
        list[dict[str, Any]] | dict[str, str] | list[Any]: JSON response with the result of the command.
    
    Aborts:
        400: If the request is not JSON or a required parameter is missing.
    """
    if not request.is_json or request.json is None:
        return abort(400, 'Request must be JSON')
    command = request.json.get('type')
    auth_method = request.json.get('user_auth', '')
    user_ident = request.json.get('user_id', '')

    if command == "check-user-auth":
        return {'is_auth_registered': user_manager.check_user_auth(auth_method)}

    if command == "get-user":
        if not (auth_method):
            return abort(400, "Missing user_auth")
        if not (user_ident):
            return abort(400, "Missing user_id")

        user = user_manager.get_user(auth_method, user_ident)
        return user.json

    if command == "list-users":
        # Get arguments for the filter by auth, admin, and banned status
        auth_filter = request.json.get('auth-filter', 'all')
        admin_filter = request.json.get('admin-filter', 'all')
        banned_filter = request.json.get('banned-filter', 'all')
        user_list = [user.json for user in user_manager.list_users(auth_filter, admin_filter, banned_filter)]
        return user_list

    elif command == 'toggle-admin':
        new_status = request.json.get('new_status')
        if new_status is None:
            return abort(400, 'Missing new_status')
        user_manager.update_admin_status(auth_method, user_ident, new_status)
        return {'response': f'Made {auth_method}.{user_ident} {"an admin" if new_status else "no longer an admin"}'}

    elif command == 'toggle-ban':
        new_status = request.json.get('new_status')
        if new_status is None:
            return abort(400, 'Missing new_status')
        user_manager.update_banned_status(auth_method, user_ident, new_status)
        return {'response': f'Made {auth_method}.{user_ident} {"banned" if new_status else "no longer banned"}'}

    elif command == 'update-requests':
        sub_action = request.json.get('action')
        if sub_action is None:
            return abort(400, 'Missing action')
        if sub_action == 'add':
            user_manager.increase_requests(auth_method, user_ident)
        elif sub_action == 'remove':
            user_manager.decrease_requests(auth_method, user_ident)
        else:
            return abort(400, f'Invalid action was given: "{sub_action}"')
        return {'response': f'Updated {auth_method}.{user_ident} requests'}

    elif command == 'remove-user':
        force_remove = request.json.get('force_remove', False)
        user_manager.remove_user_from_cache(auth_method, user_ident, force_remove=force_remove)
        return {'response': f'Removed {auth_method}.{user_ident} from the cache'}

    elif command == 'fetch-user':
        user_manager.fetch_user(auth_method, user_ident)
        return {'response': f'Fetched {auth_method}.{user_ident}'}

    elif command == 'list-cleanables':
        return user_manager.list_cleanables()

    elif command == 'clean-cache':
        user_manager.clean_cache()
        return {'response': 'Cleaned user cache'}

    else:
        return abort(400, f'Invalid command type was given: "{command}"')



"""
Module for managing chat sessions and interactions with multiple chat interfaces.

© 2024 Blaine Freestone, Carson Bush

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_logs import BaseChatLogsManager
from maeser.user_manager import User
import time
from uuid import uuid4 as uid
from langchain_community.callbacks import get_openai_callback
from langgraph.graph.graph import CompiledGraph

class ChatSessionManager:
    """
    Manages and directs sessions for multiple chat interfaces.
    """
    
    def __init__(
        self,
        chat_logs_manager: BaseChatLogsManager | None = None,
    ) -> None:
        """
        Initializes the chat session manager.

        Args:
            chat_logs_manager (BaseChatLogsManager | None): The chat logs manager to use for logging chat data.

        Returns:
            None
        """
        self.chat_logs_manager: BaseChatLogsManager | None = chat_logs_manager
        self.graphs: dict = {}

    def register_branch(self, branch_name: str, branch_label: str, graph: CompiledGraph) -> None:
        """
        Registers a branch with its information and graph.

        Args:
            branch_name (str): The name of the branch.
            branch_label (str): The label of the branch.
            graph (CompiledGraph): The graph for the branch.
        
        Returns:
            None
        """
        self.graphs[branch_name] = {
            'label': branch_label,
            'graph': graph
        }

    def get_new_session_id(self, branch_name: str, user: User | None = None) -> str:
        """
        Creates a new chat session for the given branch action.
        Includes creating a new log file for the session.

        Args:
            branch_name (str): The action of the branch to create a session for.
            user (User | None): The user to create the session for.

        Returns:
            str: The session ID for the new session.
        """
        # Generate session ID with user information if it exists
        if user:
            session_id: str = f'{uid()}-{user.auth_method}-{user.ident}'
        else:
            session_id: str = f'{uid()}-anon'

        # Create log file if chat logs manager is available
        if self.chat_logs_manager:
            self.chat_logs_manager.log(branch_name, session_id, {'user': user})

        return session_id
    
    def ask_question(self, message: str, branch_name: str, sess_id: str) -> dict:
        """
        Asks a question in a specific session of a branch.

        Args:
            message (str): The question to ask.
            branch_name (str): The action of the branch to ask the question in.
            sess_id (str): The session ID to ask the question in.

        Returns:
            dict: The response to the question.
        """
        config = {'configurable': {'thread_id': sess_id}}
        start_time = time.time()
        # Get token count for the response
        with get_openai_callback() as cb:
            response = self.graphs[branch_name]['graph'].invoke({
                'messages': [message],
            }, config=config)
            response['tokens_used'] = cb.total_tokens
            response['cost'] = cb.total_cost
        end_time = time.time()
        execution_time = end_time - start_time

        response['execution_time'] = execution_time
        
        if self.chat_logs_manager:
            self.chat_logs_manager.log(branch_name, sess_id, response)
        
        return response
    
    def add_feedback(self, branch_name: str, session_id: str, message_index: int, feedback: str) -> None:
        """
        Adds feedback to the log for a specific response in a specific session.

        Args:
            branch_name (str): The name of the branch.
            session_id (str): The session ID for the conversation.
            message_index (int): The index of the message to add feedback to.
            feedback (str): The feedback to add to the message.
        
        Returns:
            None
        """
        # Return if no chat logs manager
        if not self.chat_logs_manager:
            return
        
        self.chat_logs_manager.log_feedback(branch_name, session_id, message_index, feedback)

    def get_conversation_history(self, branch_name: str, session_id: str) -> dict:
        """
        Gets the conversation history for a specific session in a specific branch.

        Args:
            branch_name (str): The action of the branch to get the conversation history from.
            session_id (str): The session ID to get the conversation history from.

        Returns:
            dict: The conversation history for the session.
        """
        if not self.chat_logs_manager:
            return {}
        
        return self.chat_logs_manager.get_chat_history(branch_name, session_id)
    
    @property
    def branches(self) -> dict:
        """
        Returns the list of branches available for chat.

        Returns:
            dict: The list of branches available for chat.
        """
        return self.graphs
    
    @property
    def chat_log_path(self) -> str | None:
        """
        Returns the path to the logs directory.

        Returns:
            str | None: The path to the logs directory.
        """
        return self.chat_logs_manager.chat_log_path if self.chat_logs_manager else None



"""
This module contains the controller for rendering the user management page.

© 2024 Carson Bush

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import render_template
from maeser.user_manager import UserManager

def controller(
    user_manager: UserManager,
    app_name: str | None = None,
    main_logo_login: str | None = None,
    main_logo_chat: str | None = None,
    chat_head: str | None = None,
    favicon: str | None = None,
):
    """
    Render the user management page.

    Args:
        user_manager (UserManager): The user manager instance.
        app_name (str | None, optional): The name of the application. Defaults to None.
        main_logo_login (str | None, optional): The light version of the main logo. Defaults to None.
        main_logo_chat (str | None, optional): The dark version of the main logo. Defaults to None.
        chat_head (str | None, optional): The chat head image URL. Defaults to None.
        favicon (str | None, optional): The favicon image URL. Defaults to None.

    Returns:
        str: The rendered HTML for the user management page.
    """
    return render_template(
        template_name_or_list='user_management.html',
        user_manager=user_manager,
        users=user_manager.list_users(),
        main_logo_login=main_logo_login,
        main_logo_chat=main_logo_chat,
        chat_head=chat_head,
        favicon=favicon,
        app_name=app_name if app_name else 'Maeser',
        # Builtin functions not normally in Jinja templates
        len=len,
    )



"""
Module for handling feedback form submissions.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager
from flask import request, redirect

def controller(chat_sessions_manager: ChatSessionManager):
    """
    Controller function to handle the feedback form submission.

    Args:
        chat_sessions_manager (ChatSessionManager): The manager for chat sessions.

    Returns:
        Response: Redirects to the home page.
    """

    chat_logs_manager = chat_sessions_manager.chat_logs_manager

    name = request.form.get('name')
    feedback = request.form.get('feedback')
    role = request.form.get('role')
    category = request.form.get('category')

    if chat_logs_manager is not None:
        chat_logs_manager.save_feedback({
            'name': name,
            'feedback': feedback,
            'role': role,
            'category': category
        })

    return redirect('/')



"""Module for handling chat API requests.

This module contains the controller function for managing chat sessions and processing incoming messages.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager
from maeser.render import get_response_html
from flask import request, abort
from openai import RateLimitError

def controller(chat_sessions_manager: ChatSessionManager, chat_session: str):
    """Handle incoming messages for a chat session.

    Args:
        chat_sessions_manager (ChatSessionManager): The manager for chat sessions.
        chat_session (str): Chat session ID.

    Returns:
        dict: Response containing the HTML representation of the response.
    """
    posty = request.get_json()

    try:
        response = chat_sessions_manager.ask_question(posty['message'], posty['action'], chat_session)
    except RateLimitError as e:
        print(f'{type(e)}, {e}: Rate limit reached')
        abort(503, description='Rate limit reached, please try again later')
    
    return {'response': get_response_html(response['messages'][-1]), 'index': len(response['messages']) - 1}



"""
Module for handling login and GitHub OAuth2 authorization controllers.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import render_template, redirect, url_for, request, session
from flask_login import login_user, current_user
from urllib.parse import urljoin, urlparse

def controller():
    """Raises NotImplementedError for unimplemented controllers.

    This function should be replaced with a specific login controller 
    implementation, such as 'login_controller', 'github_authorize_controller', 
    or 'github_auth_callback_controller'.
    """
    raise NotImplementedError('Please import specify the login controller you want ("login_controller", "github_authorize_controller", "github_auth_callback_controller")')

def is_safe_url(target):
    """Checks if a URL is safe for redirection.

    Args:
        target (str): The target URL to check.

    Returns:
        bool: True if the URL is safe, False otherwise.
    """
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ('http', 'https') and ref_url.netloc == test_url.netloc

def login_controller(auth_manager, app_name: str | None = None, main_logo_login: str | None = None, main_logo_chat: str | None = None, favicon: str | None = None):
    """Handles user login.

    Args:
        auth_manager (AuthManager): The authentication manager to handle user authentication.

    Returns:
        Response: The response object to render the login page or redirect.
    """
    if current_user is not None and current_user.is_authenticated:
        return redirect('/')
    
    if request.method == 'POST':
        username = request.form.get('username', '')
        password = request.form.get('password', '')
        auth_method = request.form.get('authvalidator', 'invalid')
        if auth_method == 'invalid':
            return render_template(
                'login.html',
                message='Invalid Authentication Method in Request',
                main_logo_login=main_logo_login,
                main_logo_chat=main_logo_chat,
                favicon=favicon,
                app_name=app_name if app_name else "Maeser",
                authenticators=auth_manager.authenticators,
            )
        user = auth_manager.authenticate(auth_method, username, password)
        if user is None:
            return render_template(
                'login.html', 
                message='Authentication Failed',
                main_logo_login=main_logo_login,
                main_logo_chat=main_logo_chat,
                favicon=favicon,
                app_name=app_name if app_name else "Maeser",
                authenticators=auth_manager.authenticators,
            )
        if not user.is_active:
            return render_template(
                'login.html', 
                message=f'User {user.full_id_name} is Banned',
                main_logo_login=main_logo_login,
                main_logo_chat=main_logo_chat,
                favicon=favicon,
                app_name=app_name if app_name else "Maeser",
                authenticators=auth_manager.authenticators,
            )

        login_user(user)

        next_url = request.args.get('next')
        if not next_url or not is_safe_url(next_url):
            next_url = '/'

        return redirect(next_url)

    next_url = request.args.get('next')
    message = request.args.get('message', '')

    return render_template(
        'login.html', 
        message=message, 
        next=next_url,
        main_logo_login=main_logo_login,
        main_logo_chat=main_logo_chat,
        favicon=favicon,
        app_name=app_name if app_name else "Maeser",
        authenticators=auth_manager.authenticators,
    )

def github_authorize_controller(current_user, github_authenticator):
    """Handles GitHub OAuth2 authorization.

    Args:
        current_user (User): The currently logged-in user.
        github_authenticator (GitHubAuthenticator): The GitHub authenticator to get OAuth2 info.

    Returns:
        Response: The response object to redirect to the OAuth2 provider.
    """
    if not current_user.is_anonymous:
        return redirect('/')

    session['oauth2_state'], provider_url = github_authenticator.get_auth_info()
    session.modified = True
    print(f'OAuth2 state: {session["oauth2_state"]}')

    # Redirect the user to the OAuth2 provider authorization URL
    return redirect(provider_url)

def github_auth_callback_controller(current_user, auth_manager, app_name: str | None = None, main_logo_login: str | None = None, main_logo_chat: str | None = None, favicon: str | None = None, login_redirect: str = 'maeser.login'):
    if not current_user.is_anonymous:
        return redirect('/')

    # If there was an error before auth, render the login page with the error message
    if 'error' in request.args:
        print(f'An error occurred during the auth callback before authentication: {request.args}')
        error_message = request.args.get('error_description', 'Authentication failed')
        return redirect(url_for(login_redirect, message=error_message))
    
    oauth_state = session.get('oauth2_state')
    print(f'OAuth2 state at callback: {oauth_state}')

    user = auth_manager.authenticate('github', request.args, oauth_state)
    if user is None:
        return redirect(url_for(login_redirect, message='GitHub Authentication Failed'))
    if not user.is_active:
        return redirect(url_for(login_redirect, message=f'User {user.full_id_name} is Banned'))
    
    login_user(user)

    return redirect('/')



"""
Logout controller for handling user logouts and session cleanup.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask_login import logout_user
from flask import redirect, url_for, session

def controller():
    """Handles user logout and session cleanup.

    Logs out the user and removes specific session keys related to the user identity.

    Returns:
        flask.Response: A redirect response to the login page.
    """
    logout_user()
    for key in ('identity.name', 'identity.auth_type'):
        session.pop(key, None)
    return redirect(url_for('maeser.login'))


"""
This module contains the controller function to display the training form.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from flask import render_template

def controller(app_name: str | None = None, main_logo_login: str | None  = None, main_logo_chat: str | None = None, favicon: str | None = None) -> str:
    """
    Display the training form.

    Returns:
        str: Rendered training template.
    """
    role_options = ['Professor', 'Teachers Assistant']
    type_options = ['Information', 'Style']

    return render_template(
        'training.html',
        role_options=role_options,
        type_options=type_options,
        app_name=app_name if app_name else "Maeser",
        main_logo_login=main_logo_login,
        main_logo_chat=main_logo_chat,
        favicon=favicon
    )



"""
This module contains the controller function for rendering the chat logs overview page.

It handles fetching log files, applying filters, and calculating aggregate data such as total tokens and cost.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

from maeser.chat.chat_session_manager import ChatSessionManager

from flask import render_template, request


def controller(chat_sessions_manager: ChatSessionManager, app_name: str | None = None, favicon: str | None = None) -> str:
    """
    Render the home page with log files and aggregate token and cost data.

    Args:
        chat_sessions_manager (ChatSessionManager): An instance of ChatSessionManager to manage chat sessions.

    Returns:
        str: Rendered home template with log file list.
    """
    chat_logs_manager = chat_sessions_manager.chat_logs_manager
    chat_branches = chat_sessions_manager.branches

    sort_by: str = request.args.get('sort_by', 'modified')  # Default sorting by modification time
    order = request.args.get('order', 'desc')  # Default sorting order is descending
    branch_filter = request.args.get('branch', '')
    user_filter = request.args.get('user','')
    feedback_filter = request.args.get('feedback', None)

    chat_logs_overview, total_tokens, total_cost, users = chat_logs_manager.get_chat_logs_overview(sort_by, order, branch_filter, user_filter, feedback_filter) if chat_logs_manager else []
    branches = [branch for branch in chat_branches] if chat_branches else []

    return render_template(
        'chat_logs_overview.html', 
        log_files=chat_logs_overview, 
        branches=branches, 
        sort_by=sort_by, 
        order=order, 
        branch_filter=branch_filter,
        user_filter=user_filter,
        feedback_filter=feedback_filter,
        total_tokens=total_tokens, 
        total_cost=total_cost,
        users=users,
        favicon=favicon,
        app_name=app_name if app_name else "Maeser",
    )


"""
This module provides a controller for fetching the remaining requests for a user.

© 2024 Carson Bush, Blaine Freestone

This file is part of Maeser.

Maeser is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Maeser is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
Maeser. If not, see <https://www.gnu.org/licenses/>.
"""

def controller(auth_manager, current_user):
    """Fetch the remaining requests for the current user.

    Args:
        auth_manager: The authentication manager to use for fetching request data.
        current_user: The current user whose request data is being fetched.

    Returns:
        dict: A dictionary containing the number of requests remaining for the user.
    """
    return {'requests_remaining': auth_manager.get_requests_remaining(current_user.auth_method, current_user.ident)}
